#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun May 16 13:50:34 2021

@author: anton
"""

from SLiCAP import *
from values import attn, SHOW

prj = initProject('Probe project')

fileName = 'ProbeSymbolic'
#makeNetlist(fileName + '.asc', fileName)

# Create an instance of the instruction object
i1  = instruction()

# Define the circuit for this instruction and check its netlist
i1.setCircuit(fileName + '.cir')

# Display the circuit data on an HTML page
htmlPage("Circuit data")
head2html("Circuit diagram")
img2html('probe.jpg', 100)
img2html(fileName + ".svg", 500)
netlist2html(fileName + ".cir")
elementData2html(i1.circuit)
params2html(i1.circuit)

# Set the analysis type to symbolic
# Declare multiple symbolic variables:
C_a, C_b, R_a, R_b, R_s, L_g = sp.symbols("C_a C_b R_a R_b R_s L_g")

C_b_comp = C_a*R_a/R_b

i1.setSimType("symbolic")

###############################################################################
#                                                                             #
#                            Symbolic matrix equation                         #
#                                                                             #
###############################################################################

# Create an HTML page to display the results

htmlPage("MNA matrix equation")
head2html("Circuit diagram")
img2html('probe.jpg', 100)
img2html(fileName + ".svg", 500)

# Put some text on the HTML page
text2html('The MNA matrix equation of the circuit is:')

i1.setGainType('vi')
i1.setDataType('matrix')
result = i1.execute()
matrices2html(result)

###############################################################################
#                                                                             #
#         Symbolic analysis of the Laplace transfer of the gain               #
#                                                                             #
###############################################################################

# Create an HTML page to display the results

htmlPage("Symbolic Laplace Transform")
head2html("Circuit diagram")
img2html('probe.jpg', 100)
img2html(fileName + ".svg", 500)

# Define the source and the detector, and the gain type.

i1.setSource('V1')
i1.setDetector('V_scope')

# Define the source and the detector, and the gain type.

i1.setSource('V1')
i1.setDetector('V_scope')
i1.setGainType('gain')

# Laplace transform of the transfer

i1.setDataType("laplace")
result = i1.execute()

symbolic_laplace = result.laplace

text2html("The Laplace Transform of the transfer $\\frac{V_scope}{V_s}$ is:")

eqn2html("V_scope/V_s", symbolic_laplace)

# Let's see the expression for in case the probe is calibrated

text2html("If the probe is correctly calibrated, we substitute:")
eqn2html(C_b, C_a*R_a/R_b)

text2html("The expression of the transfer then simplifies to:")
# Use sympy substitution
symbolic_laplace_calibrated = symbolic_laplace.subs(C_b, C_b_comp)
eqn2html("V_scope/V_s", symbolic_laplace_calibrated)

text2html("If we factorize this result we obtain:")
# Use sympy factorization
symbolic_laplace_calibrated = sp.factor(symbolic_laplace_calibrated)
eqn2html("V_scope/V_s", symbolic_laplace_calibrated)

text2html("If we normalize this result, we obtain:")
# Use the SLiCAP "normalizeRational" function to see the DC transfer and the 
# frequency-dependent part
symbolic_laplace_calibrated = normalizeRational(symbolic_laplace_calibrated)
eqn2html("V_scope/V_s", normalizeRational(symbolic_laplace_calibrated))

###############################################################################
#                                                                             #
#            Initial value and final value of the unit-step response          #
#                                                                             #
###############################################################################

# If we ignore the source resistance, the final value of the step response
# is determined by the ratio of two resistors. We can evaluate this by taking
# R_s = 0 and calculate: lim s-->0 of the Laplace Transform.

symbolic_laplace_no_Rs = symbolic_laplace.subs(R_s, 0)
final_value = sp.limit(symbolic_laplace_no_Rs, ini.Laplace, '0')

# If we ignore the bandwidth limitation, the initial value of the step response
# is determined by the ratio of the capacitors. We can evaluate this by taking
# R_s = 0 and L_s = 0 and calculate: lim s-->oo of the Laplace Transform.

high_bw_laplace = symbolic_laplace.subs({R_s: 0, L_g: 0})
initial_value = sp.simplify(sp.limit(high_bw_laplace, ini.Laplace,'oo'))

text2html("The initial value ($R_s=0, \\, L_g=0$) and the final value " +
          "($R_s=0$) of the step response, are:")
eqn2html('mu(0)', initial_value)
eqn2html('mu(oo)', final_value)

text2html("The probe is calibrated if the initial value equals the final " +
          "value:")
cal_value = sp.solve(initial_value - final_value, C_b)[0]

eqn2html(C_b, cal_value)

###############################################################################
#                                                                             #
#                     Overshoot as a function of the detuning                 #
#                                                                             #
###############################################################################

# Now that we have the initial value and the final value, we can express the
# Overshoot (or undershoot) in terms of the ralative detuning of C_b

# Define the relative detuning of C_b

delta = sp.Symbol("delta")

# Calculate the overshoot as a function of delta
epsilon = initial_value.subs(C_b, (1 + delta)*C_b_comp) - final_value
# Let us express the resitance R_a in R_b and the probe attenuation factor

R_attn = (attn-1)*R_b
# Simplify the expression and account for the attenuation
epsilon = sp.simplify(attn*epsilon.subs(R_a, R_attn))

text2html("The relative overshoot $\\epsilon$ can be expressed in terms of " +
          "the relative detuning $\\delta$ of $C_b$:")
eqn2html('epsilon', epsilon)
# Plot this function
# One way of doing this is define epsilon and delta as circuit parameters
# and use 'params' as data type

i1.defPar('delta', 0)
i1.defPar('epsilon', epsilon)

i1.setDataType('params')
result = i1.execute()
ovrsht_det = plotSweep('overshoot_detuning', 
                       'relative overshoot as a function of the relative ' +
                       'detuning', result, -0.25, 0.25, 100, sweepVar ='delta', 
                       yVar='epsilon', xVar = 'delta', funcType = 'param',
                       yUnits = '-', xUnits = '-', show = SHOW)

# Place the plot on the html page

fig2html(ovrsht_det, 600, caption='Relative overshoot versus the relative' +
         'detuning of the probe')

###############################################################################
#                                                                             #
#      Symbolic pole-zero analysis of the Laplace transfer of the gain        #
#                                                                             #
###############################################################################

# Create an HTML page to display the results

htmlPage("Symbolic Pole-zero analysis")
head2html("Circuit diagram")
img2html('probe.jpg', 100)
img2html(fileName + ".svg", 500)

# This MAY work for simple functions

i1.setDataType('pz')
result = i1.execute()

pz2html(result)

head2html("Symbolic Pole-zero analysis after compensation")

# For this, you need to redefine the value of C2. A function for doing this
# has not been implemented because element values are only checked when
# checking  the netlist. But below is the method that works, but be careful!
# In this way you can change the type of the element (R, L, C), or make errors
# in expressions that are not checked.

i1.circuit.elements['C2'].params['value'] = C_b_comp
# Execute symbolic pole-zero analysis
result = i1.execute()

# Print the result on the HTML page
pz2html(result)

###############################################################################
#                                                                             #
#                         Symbolic step response                              #
#                                                                             #
###############################################################################

# Create an HTML page to display the results

htmlPage("Symbolic time-domain analysis")
head2html("Circuit diagram")
img2html('probe.jpg', 100)
img2html(fileName + ".svg", 500)

# This MAY work for VERY simple functions

# The inverse Laplace Transform is performed by maxima. Maxima is called by
# Python in a subprocess. Questions asked by maxima cannot be answered
# by Python. Symbolic calculation of the Inverse Laplace Transform often
# requires such Q&A communication and results in a time-out error.
# In the future interactive operation with maxima will be made possible using
# socket communication.
 

i1.setDataType('step') 

# 1st-order detuned, R_s = 0
i1.circuit.elements['L1'].params['value'] = 0
i1.circuit.elements['R1'].params['value'] = 0
i1.circuit.elements['C2'].params['value'] = C_b

result = i1.execute()
text2html('The 1-rst order ($R_s=0$) detuned unit step response is found as:')
eqn2html('mu_t', result.stepResp)
stepDetuned = result.stepResp

# 1st-order tuned R_s != 0
i1.circuit.elements['L1'].params['value'] = 0
i1.circuit.elements['R1'].params['value'] = R_s
i1.circuit.elements['C2'].params['value'] = C_b_comp

result = i1.execute()
text2html('The 1-rst order ($R_s \\neq 0,\\, C_b=C_a \\frac{R_a}{R_b}$) ' +
          'tuned unit step response is found as:')
eqn2html('mu_t', result.stepResp)

###############################################################################
#                                                                             #
#                   Settling time versus detuning                             #
#                                                                             #
###############################################################################

n = sp.Symbol('n') # Number of bits

DETUNING = ['positive', 'negative']     # We need to runs this simulation twice

i1.defPar('n', 8)                       # Number of bits
i1.defPar('delta', 0)                   # Relative calibration error of C_b
i1.defPar('tau_s', 0)                   # Settling time

# Step the number of bits for determination of the settling error
i1.setStepVar('n')
i1.setStepStart(8)
i1.setStepStop(12)
i1.setStepNum(5)
i1.setStepMethod('lin')
i1.stepOn()

for detune in DETUNING:
    if detune == 'positive':
        settling_error = -final_value*2**(-n)
        dMin = 0.005
        dMax = 0.1
    else:
        settling_error = final_value*2**(-n)
        dMax = -0.005
        dMin = -0.1
        
    settling_error_t = sp.simplify(stepDetuned - final_value)
    text2html('The settling error to the final value of a wrongly tuned ' +
              'probe can be written as a function of time:')
    eqn2html('epsilon_t',settling_error_t)
    
    t_settling = sp.solve(settling_error_t - settling_error, sp.Symbol('t'))[0]
    t_settling_detuning = t_settling.subs({C_b: (1 + delta)*C_b_comp, 
                                           R_a: R_attn})
    t_settling_detuning = sp.simplify(t_settling_detuning)
    
    text2html('After equating this value with a 1 LSB error, we find the ' +
              'settling time as function of the relative ' + detune +
              'detuning $\\delta$ of $C_b$:')
    eqn2html('tau_s', t_settling_detuning)
    i1.defPar('tau_s', t_settling_detuning)
    i1.setDataType('params')
    result = i1.execute()
    settle_det = plotSweep('settling_detuning_' + detune, 
                           'settling time as a function of the relative ' +
                           detune + ' detuning', result, dMin, dMax, 500, 
                           sweepVar ='delta', yVar='tau_s', xVar = 'delta', 
                           funcType = 'param', yUnits = 's', yScale='u', 
                           xUnits = '-', show = SHOW)
    
    fig2html(settle_det, 600)